/*

Copyright (c) 2004, Roland J. Weiss.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, this
    list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
  * Neither the name of the University of Tuebingen nor the names of its
    contributors may be used to endorse or promote products derived from this
    software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

import java.lang.IllegalArgumentException;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

import java.util.TreeSet;
import java.util.Collection;

class StringExpander
{
  /**
  This method takes a String s and expands the contained escape characters
  according to the following explained in the pdwrap manual. The expanded
  string is returned.
  */
  static public String expand(String s)
  {
    // This String accumulates the expanded string.
    String expanded = new String();

    // Fill escape char collections.
    TreeSet simple_escapes = new TreeSet();
    copyCharsToCollection("belgpf()%", simple_escapes);

    TreeSet arg_escapes = new TreeSet();
    copyCharsToCollection("?=", arg_escapes);

    // The top-level scanner just looks for the percent escape char.
    Pattern p_percent = Pattern.compile("%");

    // These vars handle which section of the string are accumulated and where
    // processing continues.
    // Before entering the main loop, they both point to the next character that
    // should be processed. In the main loop, end will serve as marker to the
    // current position of interest.
    int start = 0;
    int end = 0;

    // Return empty string for empty (sub)strings.
    if (s.length() == start) return "";

    // The matcher is initialized with the unexpanded string.
    Matcher m = p_percent.matcher(s);
    // This flag indicates if a percent char was found.
    boolean found;

    // This is the main loop. It works as follows:
    // (1) Look for the percent char.
    // (2) Accumulate the substring in front of this char, or the whole string.
    // (3) If a percent char is found, expand the command.
    // (4) If no percent char was found we are done.
    do {
      
      // Start by scanning the string from the current position for the % char.
      found = m.find(start);
      if (!found)
      {
        end = s.length(); // Point end marker to end of the string.
      }
      else
      {
        end = m.end()-1; // Point end marker to percent sign.
      }
      // Accumulate substring.
      expanded += s.substring(start, end);

      // Expand command.
      if (found)
      {
        // Advance end marker to command character.
        ++end;
        char command_char = s.charAt(end);

        // Var for balanced brace indices.
        TextDelimiter delim = new TextDelimiter();

        // Simple escapes are replaced by the corresponding text.
        if (simple_escapes.contains(new Character(command_char)))
        {
          //System.out.println(" command char found: " + command_char);
          expanded += expandCommandChar(command_char);
          ++end;
        }
        // Conditional escapes are followed by text strings in braces.
        else if (arg_escapes.contains(new Character(command_char)))
        {
          //System.out.println(" command char found: " + command_char);

          // Here, we check if the expanded text corresponding to the command
          // character is not empty. If so, expand the first brace text,
          // otherwise simply skip it.
          if (command_char == '?')
          {
            // Advance end marker to next command character.
            ++end;
            String checkExpansion = expandCommandChar(s.charAt(end));
            // Advance end marker to open brace.
            ++end;

            getBraceText(s, end, delim);
            String braceText = s.substring(delim.getFirst()+1, delim.getLast());
            //System.out.println(" first: " + delim.getFirst());
            //System.out.println(" last : " + delim.getLast());
            //System.out.println(" brace text: " + braceText);

            if (checkExpansion.length() > 0)
            {
              expanded += expand(braceText);
            }
            // Set end marker just behind brace.
            end = delim.getLast()+1;
          }
          // Here, we check if the expanded text corresponding to the command
          // characters is equal. If so, expand the first brace text, otherwise
          // the second one.
          else if (command_char == '=')
          {
            // Advance end marker to next command characters.
            ++end;
            String checkExpansion = expandCommandChar(s.charAt(end));
            ++end;
            String checkExpansion2 = expandCommandChar(s.charAt(end));
            // Advance end marker to open brace.
            ++end;

            if (checkExpansion.equals(checkExpansion2))
            {
              // Get content of first brace.
              getBraceText(s, end, delim);
              String braceText = s.substring(delim.getFirst()+1, delim.getLast());
              //System.out.println(" first: " + delim.getFirst());
              //System.out.println(" last : " + delim.getLast());
              //System.out.println(" brace text: " + braceText);
              expanded += expand(braceText);
              end = delim.getLast()+1;

              // Skip second brace.
              getBraceText(s, end, delim);
              end = delim.getLast()+1;
            }
            else
            {
              // Skip first brace.
              getBraceText(s, end, delim);
              end = delim.getLast()+1;

              // Get content of second brace.
              getBraceText(s, end, delim);
              String braceText = s.substring(delim.getFirst()+1, delim.getLast());
              //System.out.println(" first: " + delim.getFirst());
              //System.out.println(" last : " + delim.getLast());
              //System.out.println(" brace text: " + braceText);
              expanded += expand(braceText);
              end = delim.getLast()+1;
            }
          }
        }
        else
        {
          throw new IllegalArgumentException("Unkown command character found: " +
                                             command_char);
        }
        // Adjust markers for next loop.
        start = end;
      }

    } while (found);

    // Finally, return the accumulated string.
    return expanded;
  }

  /**
  Find the indices of the given String s that correspond to the matching
  */
  static void getBraceText(String s, int start, TextDelimiter delim)
  {
    int brace_level = 0;

    Pattern pBracesPercent = Pattern.compile("[%()]");
    Matcher m = pBracesPercent.matcher(s);

    boolean found = m.find(start);
    if (!found)
    {
      throw new IllegalArgumentException("Opening brace expected, but no special character present.");
    }
    int next = m.end();
    if (s.charAt(next-1) != '(')
    {
      throw new IllegalArgumentException("String starts with illegal special character: " + s.charAt(next-1));
    }
    // Point <first> to opening brace.
    delim.setFirst(next-1);

    do
    {
      // Look for the next special character.
      found = m.find(next);
      if (!found)
      {
        throw new IllegalArgumentException("Missing properly matching closing brace.");
      }
      else next = m.end();

      // Get the detected character.
      char special_char = s.charAt(next-1);

      if (special_char == '%')
      {
        // Skip next character, this will ignore escaped braces.
        ++next;
        continue;
      }

      // Handle closing brace.
      if (special_char == ')')
      {
        // This is the matching closing brace.
        if (brace_level == 0)
        {
          delim.setLast(next-1);
          return;
        }
        else if (brace_level < 0) // Seems this can never happen?!
        {
          throw new IllegalArgumentException("Too many/unbalanced closing braces in string.");
        }
        else 
        {
          --brace_level;
        }
      }
      // Handle opening brace.
      else if (special_char == '(')
      {
        ++brace_level;
      }
    } while (true);
  }

  /**
  Return text string that corresponds to the given command character.
  */
  static String expandCommandChar(char command_char)
  {
    if (command_char == 'l')
    {
      return label;
    }
    else if (command_char == 'g')
    {
      return genLabel;
    }
    else if (command_char == 'r')
    {
      return genRef;
    }
    else if (command_char == 'b')
    {
      return Integer.toString(begin);
    }
    else if (command_char == 'e')
    {
      return Integer.toString(end);
    }
    else if (command_char == 'p')
    {
      return progLang;
    }
    else if (command_char == 'f')
    {
      return fileName;
    }
    else if ((command_char == '(') || (command_char == ')') || (command_char == '%'))
    {
      return new String(new Character(command_char).toString());
    }
    else 
    {
      return "ERROR: UNKNOWN COMMAND CHARACTER [" + command_char + "]";
    }
  }

  /**
  Copy every character from String s to collection c as Character object.
  */
  static void copyCharsToCollection(String s, Collection c)
  {
    for (int i = 0; i < s.length(); ++i)
    {
      c.add(new Character(s.charAt(i)));
    };
  }

  /* Methods to update the static properties used in the expand method. */
  static void setFileName(String s) { fileName = s; }
  static void setProgLang(String s) { progLang = s; }
  static void setLabel(String s) { label = s; }
  static void setGenLabel(String s) { genLabel = s; }
  static void setGenRef(String s) { genRef = s; }
  static void setBegin(int i) { begin = i; }
  static void setEnd(int i) { end = i; }

  /* Static properties used in the expand method. */
  static String fileName = new String();
  static String progLang = new String();
  static int begin = 0;
  static int end = 0;
  static String label = new String();
  static String genLabel = new String();
  static String genRef = new String();
  
} // StringExpander

/**
Simple class that holds two interger indices that delimit a text section.
Used by class StringExpander.
*/
class TextDelimiter
{
  TextDelimiter()
  {
    first = 0;
    last = 0;
  }

  int getFirst() { return first; }
  int getLast() { return last; }
  void setFirst(int i) { first = i; }
  void setLast(int i) { last = i; }
  
  int first, last;
} // TextDelimiter
